@@ -1,7 +1,40 @@
+from __future__ import annotations
+
+from typing import Any
+from typing import Callable
+from typing import ContextManager
+from typing import Dict
+from typing import List
+from typing import Optional
+from typing import overload
+from typing import TextIO
+from typing import Tuple
+from typing import TYPE_CHECKING
+from typing import Union
+
+from .migration import _ProxyTransaction
 from .migration import MigrationContext
 from .. import util
 from ..operations import Operations
 
+if TYPE_CHECKING:
+    from typing import Literal
+
+    from sqlalchemy.engine.base import Connection
+    from sqlalchemy.sql.elements import ClauseElement
+    from sqlalchemy.sql.schema import MetaData
+
+    from ..config import Config
+    from ..ddl import DefaultImpl
+    from ..operations.ops import MigrateOperation
+    from ..script.base import ScriptDirectory
+
+_RevNumber = Optional[Union[str, Tuple[str, ...]]]
+
+ProcessRevisionDirectiveFn = Callable[
+    [MigrationContext, Tuple[str, str], List["MigrateOperation"]], None
+]
+
 
 class EnvironmentContext(util.ModuleClsProxy):
 
@@ -66,21 +99,23 @@ class EnvironmentContext(util.ModuleClsProxy):
 
     """
 
-    _migration_context = None
+    _migration_context: Optional[MigrationContext] = None
 
-    config = None
+    config: Config = None  # type:ignore[assignment]
     """An instance of :class:`.Config` representing the
     configuration file contents as well as other variables
     set programmatically within it."""
 
-    script = None
+    script: ScriptDirectory = None  # type:ignore[assignment]
     """An instance of :class:`.ScriptDirectory` which provides
     programmatic access to version files within the ``versions/``
     directory.
 
     """
 
-    def __init__(self, config, script, **kw):
+    def __init__(
+        self, config: Config, script: ScriptDirectory, **kw: Any
+    ) -> None:
         r"""Construct a new :class:`.EnvironmentContext`.
 
         :param config: a :class:`.Config` instance.
@@ -94,7 +129,7 @@ class EnvironmentContext(util.ModuleClsProxy):
         self.script = script
         self.context_opts = kw
 
-    def __enter__(self):
+    def __enter__(self) -> EnvironmentContext:
         """Establish a context which provides a
         :class:`.EnvironmentContext` object to
         env.py scripts.
@@ -106,15 +141,15 @@ class EnvironmentContext(util.ModuleClsProxy):
         self._install_proxy()
         return self
 
-    def __exit__(self, *arg, **kw):
+    def __exit__(self, *arg: Any, **kw: Any) -> None:
         self._remove_proxy()
 
-    def is_offline_mode(self):
+    def is_offline_mode(self) -> bool:
         """Return True if the current migrations environment
         is running in "offline mode".
 
         This is ``True`` or ``False`` depending
-        on the the ``--sql`` flag passed.
+        on the ``--sql`` flag passed.
 
         This function does not require that the :class:`.MigrationContext`
         has been configured.
@@ -136,10 +171,10 @@ class EnvironmentContext(util.ModuleClsProxy):
         """
         return self.get_context().impl.transactional_ddl
 
-    def requires_connection(self):
+    def requires_connection(self) -> bool:
         return not self.is_offline_mode()
 
-    def get_head_revision(self):
+    def get_head_revision(self) -> _RevNumber:
         """Return the hex identifier of the 'head' script revision.
 
         If the script directory has multiple heads, this
@@ -154,7 +189,7 @@ class EnvironmentContext(util.ModuleClsProxy):
         """
         return self.script.as_revision_number("head")
 
-    def get_head_revisions(self):
+    def get_head_revisions(self) -> _RevNumber:
         """Return the hex identifier of the 'heads' script revision(s).
 
         This returns a tuple containing the version number of all
@@ -166,7 +201,7 @@ class EnvironmentContext(util.ModuleClsProxy):
         """
         return self.script.as_revision_number("heads")
 
-    def get_starting_revision_argument(self):
+    def get_starting_revision_argument(self) -> _RevNumber:
         """Return the 'starting revision' argument,
         if the revision was passed using ``start:end``.
 
@@ -195,7 +230,7 @@ class EnvironmentContext(util.ModuleClsProxy):
                 "No starting revision argument is available."
             )
 
-    def get_revision_argument(self):
+    def get_revision_argument(self) -> _RevNumber:
         """Get the 'destination' revision argument.
 
         This is typically the argument passed to the
@@ -213,7 +248,7 @@ class EnvironmentContext(util.ModuleClsProxy):
             self.context_opts["destination_rev"]
         )
 
-    def get_tag_argument(self):
+    def get_tag_argument(self) -> Optional[str]:
         """Return the value passed for the ``--tag`` argument, if any.
 
         The ``--tag`` argument is not used directly by Alembic,
@@ -233,7 +268,21 @@ class EnvironmentContext(util.ModuleClsProxy):
         """
         return self.context_opts.get("tag", None)
 
-    def get_x_argument(self, as_dictionary=False):
+    @overload
+    def get_x_argument(  # type:ignore[misc]
+        self, as_dictionary: Literal[False] = ...
+    ) -> List[str]:
+        ...
+
+    @overload
+    def get_x_argument(  # type:ignore[misc]
+        self, as_dictionary: Literal[True] = ...
+    ) -> Dict[str, str]:
+        ...
+
+    def get_x_argument(
+        self, as_dictionary: bool = False
+    ) -> Union[List[str], Dict[str, str]]:
         """Return the value(s) passed for the ``-x`` argument, if any.
 
         The ``-x`` argument is an open ended flag that allows any user-defined
@@ -282,34 +331,36 @@ class EnvironmentContext(util.ModuleClsProxy):
 
     def configure(
         self,
-        connection=None,
-        url=None,
-        dialect_name=None,
-        dialect_opts=None,
-        transactional_ddl=None,
-        transaction_per_migration=False,
-        output_buffer=None,
-        starting_rev=None,
-        tag=None,
-        template_args=None,
-        render_as_batch=False,
-        target_metadata=None,
-        include_name=None,
-        include_object=None,
-        include_schemas=False,
-        process_revision_directives=None,
-        compare_type=False,
-        compare_server_default=False,
-        render_item=None,
-        literal_binds=False,
-        upgrade_token="upgrades",
-        downgrade_token="downgrades",
-        alembic_module_prefix="op.",
-        sqlalchemy_module_prefix="sa.",
-        user_module_prefix=None,
-        on_version_apply=None,
-        **kw
-    ):
+        connection: Optional[Connection] = None,
+        url: Optional[str] = None,
+        dialect_name: Optional[str] = None,
+        dialect_opts: Optional[Dict[str, Any]] = None,
+        transactional_ddl: Optional[bool] = None,
+        transaction_per_migration: bool = False,
+        output_buffer: Optional[TextIO] = None,
+        starting_rev: Optional[str] = None,
+        tag: Optional[str] = None,
+        template_args: Optional[Dict[str, Any]] = None,
+        render_as_batch: bool = False,
+        target_metadata: Optional[MetaData] = None,
+        include_name: Optional[Callable[..., bool]] = None,
+        include_object: Optional[Callable[..., bool]] = None,
+        include_schemas: bool = False,
+        process_revision_directives: Optional[
+            ProcessRevisionDirectiveFn
+        ] = None,
+        compare_type: bool = False,
+        compare_server_default: bool = False,
+        render_item: Optional[Callable[..., bool]] = None,
+        literal_binds: bool = False,
+        upgrade_token: str = "upgrades",
+        downgrade_token: str = "downgrades",
+        alembic_module_prefix: str = "op.",
+        sqlalchemy_module_prefix: str = "sa.",
+        user_module_prefix: Optional[str] = None,
+        on_version_apply: Optional[Callable[..., None]] = None,
+        **kw: Any,
+    ) -> None:
         """Configure a :class:`.MigrationContext` within this
         :class:`.EnvironmentContext` which will provide database
         connectivity and other configuration to a series of
@@ -789,7 +840,7 @@ class EnvironmentContext(util.ModuleClsProxy):
             opts=opts,
         )
 
-    def run_migrations(self, **kw):
+    def run_migrations(self, **kw: Any) -> None:
         """Run migrations as determined by the current command line
         configuration
         as well as versioning information present (or not) in the current
@@ -809,10 +860,15 @@ class EnvironmentContext(util.ModuleClsProxy):
         first been made available via :meth:`.configure`.
 
         """
+        assert self._migration_context is not None
         with Operations.context(self._migration_context):
             self.get_context().run_migrations(**kw)
 
-    def execute(self, sql, execution_options=None):
+    def execute(
+        self,
+        sql: Union[ClauseElement, str],
+        execution_options: Optional[dict] = None,
+    ) -> None:
         """Execute the given SQL using the current change context.
 
         The behavior of :meth:`.execute` is the same
@@ -826,7 +882,7 @@ class EnvironmentContext(util.ModuleClsProxy):
         """
         self.get_context().execute(sql, execution_options=execution_options)
 
-    def static_output(self, text):
+    def static_output(self, text: str) -> None:
         """Emit text directly to the "offline" SQL stream.
 
         Typically this is for emitting comments that
@@ -837,7 +893,9 @@ class EnvironmentContext(util.ModuleClsProxy):
         """
         self.get_context().impl.static_output(text)
 
-    def begin_transaction(self):
+    def begin_transaction(
+        self,
+    ) -> Union[_ProxyTransaction, ContextManager]:
         """Return a context manager that will
         enclose an operation within a "transaction",
         as defined by the environment's offline
@@ -883,7 +941,7 @@ class EnvironmentContext(util.ModuleClsProxy):
 
         return self.get_context().begin_transaction()
 
-    def get_context(self):
+    def get_context(self) -> MigrationContext:
         """Return the current :class:`.MigrationContext` object.
 
         If :meth:`.EnvironmentContext.configure` has not been
@@ -895,7 +953,7 @@ class EnvironmentContext(util.ModuleClsProxy):
             raise Exception("No context has been configured yet.")
         return self._migration_context
 
-    def get_bind(self):
+    def get_bind(self) -> Connection:
         """Return the current 'bind'.
 
         In "online" mode, this is the
@@ -906,7 +964,7 @@ class EnvironmentContext(util.ModuleClsProxy):
         has first been made available via :meth:`.configure`.
 
         """
-        return self.get_context().bind
+        return self.get_context().bind  # type: ignore[return-value]
 
-    def get_impl(self):
+    def get_impl(self) -> DefaultImpl:
         return self.get_context().impl
